
#include "generator.h"
#include "ui.h"
#include "uiplot.h"

#include "vdp.h"
#include "gensound.h"
#include "cpu68k.h"
#include "mem68k.h"
#include "cpuz80.h"
#include "event.h"
#include "state.h"

// General Video options:

// This is used to enable the 64bit transfer of the txr
// to the TA from the internal array (palette expanded).
#define USE_64_BIT_TXR_XFER 1
//#undef USE_64_BIT_TXR_XFER

// Do not use a twiddled txr as it is really slow.
// I am going to work on speeding this up soon.
#undef USE_TWIDDLED_TXR

char *ui_initload = NULL;
uint8 ui_vdpsimple = 0;
uint8 auto_frameskip = 0;
uint8 auto_frameskip_count = 4;
uint8 ui_no_plot = 0;
pvr_ptr_t txr_base;
pvr_ptr_t txr_base_ex;
uint8 screen[512*512];
uint16 screen_ex[512*512];
pvr_poly_cxt_t txr_ex_ctx;
pvr_poly_cxt_t txr_pal_ctx;
#define PAL_BANK_SHIFT  (25)
#define PAL_BANK_MASK (3 << PAL_BANK_SHIFT)

unsigned int (*palette)[4][256] = (unsigned int (*)[4][256])0xa05f9000;
unsigned short pals[256] = {0};

enum
{
	PALETTED_TXR, EXPANDED_TXR,
} ui_rendermode = EXPANDED_TXR;

typedef struct
{
  uint8 address;
  uint8 vmus[6];
  uint8 vmu_count;
} t_controller;

t_controller controllers[4];

void ui_initjoysticks(void)
{
  int idx;
  uint32 unit;
  int controller_count = 0;

  for(idx = 0; idx < 4; idx++)
  {
    controllers[idx].address = 0;
    for(unit = 0; unit < 6; unit++)
    {
      controllers[idx].vmus[unit] = 0;
    }
    controllers[idx].vmu_count = 0;

    for(unit = 0; unit < 6; unit++)
    {
      if(maple_device_func(idx, unit) & MAPLE_FUNC_CONTROLLER)
      {
        controllers[idx].address = maple_addr(idx, unit);
	controller_count++;
      }
      else if(maple_device_func(idx, unit) & MAPLE_FUNC_LCD)
      {
        controllers[idx].vmus[controllers[idx].vmu_count++] = maple_addr(idx, unit);
      }
    }
  }
}

int ui_init(int argc, char *argv[])
{
	txr_base = (pvr_ptr_t)pvr_mem_malloc(512*512);
	txr_base_ex = (pvr_ptr_t)pvr_mem_malloc(512*512*2);

	pvr_poly_cxt_txr(&txr_pal_ctx, PVR_LIST_TR_POLY, PVR_TXRFMT_PAL8BPP|PVR_TXRFMT_TWIDDLED, 512, 512, txr_base, 0);
#ifdef USE_TWIDDLED_TXR
	pvr_poly_cxt_txr(&txr_ex_ctx, PVR_LIST_TR_POLY, PVR_TXRFMT_RGB565|PVR_TXRFMT_TWIDDLED, 512, 512, txr_base_ex, 0);
#else
	pvr_poly_cxt_txr(&txr_ex_ctx, PVR_LIST_TR_POLY, PVR_TXRFMT_RGB565|PVR_TXRFMT_NONTWIDDLED, 512, 512, txr_base_ex, 0);
#endif
	ui_initjoysticks();
	return 0;
}

void ui_final()
{
	pvr_mem_free(txr_base);
	pvr_mem_free(txr_base_ex);
}

int ui_check_cond(uint32 flags)
{
	int i;
	cont_cond_t cond;
	for(i = 0; i < 4; i++)
	{
		if(controllers[i].address == 0)
		{
			continue;
		}
		cont_get_cond(controllers[i].address, &cond);
		if(!(cond.buttons & flags))
		{
			return 1;
		}
	}
	return 0;
}

void ui_updatepads(void)
{
  int i;
  int pad = 0;
  cont_cond_t cond;

  for(i = 0; i < 4; i++)
  {
    if(controllers[i].address == 0)
    {
      continue;
    }
    cont_get_cond(controllers[i].address, &cond);
    mem68k_cont[pad].up = !(cond.buttons & CONT_DPAD_UP);
    mem68k_cont[pad].right = !(cond.buttons & CONT_DPAD_RIGHT);
    mem68k_cont[pad].down = !(cond.buttons & CONT_DPAD_DOWN);
    mem68k_cont[pad].left = !(cond.buttons & CONT_DPAD_LEFT);
    mem68k_cont[pad].start = !(cond.buttons & CONT_START);
    mem68k_cont[pad].a = !(cond.buttons & CONT_X);
    mem68k_cont[pad].b = !(cond.buttons & CONT_A);
    mem68k_cont[pad].c = !(cond.buttons & CONT_B);
    pad++;

    if(cond.ltrig && !cond.rtrig)
    {
      ui_vdpsimple = !ui_vdpsimple;
    }
    else if(!cond.ltrig && cond.rtrig)
    {
      sound_stop();
      gen_reset();
      sound_start();
    }
    else if(cond.rtrig && cond.ltrig && !gen_quit)
    {
      gen_quit = 1;
    }
  }
}

#define USEC_FRAME_NTSC 16667
#define USEC_FRAME_PAL
#define USEC_FRAME (vdp_pal ? USEC_FRAME_PAL : USEC_FRAME_NTSC)

struct timeval
{
	long tv_sec;
	long tv_usec;
};

int ui_loop()
{
	struct timeval oldclk, newclk;
	uint8 frames_to_do;
	uint32 usec = 0;

	if(sound_on || !auto_frameskip)
	{
		ui_no_plot = 0; 
		// With sound enabled we do not need to worry about
		// frameskip as it is internally enabled by the frame
		// event runner.
		while(!gen_quit)
		{
			event_doframe();
			ui_updatepads();
		}
	}
	else
	{
		// Deal with auto-frameskip
		gettimeofday(&oldclk, NULL);
		while(!gen_quit)
		{
			//
			gettimeofday(&newclk, NULL);
			if(newclk.tv_usec < oldclk.tv_usec)
			{
				usec += 1000000 + newclk.tv_usec - oldclk.tv_usec;
			}
			else
			{
				usec += newclk.tv_usec - oldclk.tv_usec;
			}
			frames_to_do = usec / USEC_FRAME;
			usec %= USEC_FRAME;
			oldclk = newclk;

			frames_to_do %= auto_frameskip_count;
			for(; frames_to_do > 1; --frames_to_do)
			{
				ui_no_plot = 1; 
				event_doframe();
				ui_updatepads();
			}
			ui_no_plot = 0; 
			event_doframe();
			ui_updatepads();
		}
	}
	gen_quit = 0;
	return 0;
}

void inline palette_update()
{
	int i;
	uint8 *p = (uint8 *)vdp_cram;
	for(i = 0; i < 64; i++)
	{
		p = (uint8 *)(vdp_cram + (i << 1));
		pals[i] =          /* normal */
			(p[0] & 0x0E) << 1 |
			(p[1] & 0x0E) << 12 |
			(p[1] & 0xE0) << 3;
		pals[i + 64] =     /* hilight */
			(p[0] & 0x0E) |
			(p[1] & 0x0E) << 11|
			(p[1] & 0xE0) << 2|
			(16) | (16 << 11) |
			(16 << 6);
		pals[i + 128] =    /* shadow */
			(p[0] & 0x0E) |
			(p[1] & 0x0E) << 11|
			(p[1] & 0xE0) << 2;
		vdp_cramf[i] = 0;
	}
	if(ui_rendermode != PALETTED_TXR) return;
	for(i = 0; i < 64; i++)
	{
		if(vdp_cramf[i])
		{
			vdp_cramf[i] = 0;
			(*palette)[0][i      ] = (pals[i      ] & 0x7FFF);
			(*palette)[0][i + 64 ] = (pals[i + 64 ] & 0x7FFF);
			(*palette)[0][i + 128] = (pals[i + 128] & 0x7FFF);
		}
	}
}

#define QACR0 (*(volatile unsigned int *)(void *)0xff000038)
#define QACR1 (*(volatile unsigned int *)(void *)0xff00003c)

#define COPYPIXEL(n) \
{			\
  unsigned short _tmp = pals[*s++];		\
  d[n] = _tmp|(pals[*s++]<<16);			\
}

static void texture_memcpy64_pal(void *dest, void *src)
{
	int cnt = 64;
  unsigned char *s = (unsigned char *)src;
  unsigned int *d = (unsigned int *)(void *)
    (0xe0000000 | (((unsigned long)dest) & 0x03ffffe0));
  QACR0 = (((unsigned int)dest>>26)<<2)&0x1c;
  QACR1 = (((unsigned int)dest>>26)<<2)&0x1c;
  while(cnt--) {
    COPYPIXEL(0);
    COPYPIXEL(1);
    COPYPIXEL(2);
    COPYPIXEL(3);
    //asm("pref @%0" : : "r" (s+4*16));
    COPYPIXEL(4);
    COPYPIXEL(5);
    COPYPIXEL(6);
    COPYPIXEL(7);
    asm("pref @%0" : : "r" (d));
    d += 8;
  }
}

void ui_line(int line)
{
	uint8 *p;

	if(ui_vdpsimple || ui_no_plot)
	{
		return;
	}
	if(line < 0 || line >= (int)vdp_vislines)
	{
		return;
	}
	p = &screen[0] + (line * 512);
	if((vdp_reg[12] & 0x06) == 0x06)
	{
		vdp_renderline(line, p, vdp_oddframe);
	}
	else
	{
		vdp_renderline(line, p, 0);
	}
	palette_update();
}

void ui_endfield(void)
{
	pvr_vertex_t vert;
	pvr_poly_hdr_t screen_poly;
	uint32 y_size = 480;
	uint32 x_size = 640;
	uint32 x = 0, y = 0;
	uint32 u_max = (vdp_reg[12] & 1) ? 320 : 256;

	if(ui_no_plot)
	{
		return;
	}
	palette_update();
	
	if(ui_vdpsimple)
	{
		vdp_renderframe(&screen[0], 512);
	}
	// on a VGA cable we need to offset by 40.
	if(!vid_check_cable()) // return value of 0 means VGA
	{
		y += 40;
	}
	
	if(ui_rendermode == PALETTED_TXR)
	{
		txr_twiddle_copy_general(&screen[0], txr_base, 512, 512, 8);
		pvr_poly_compile(&screen_poly, &txr_pal_ctx);
	}
	else
	{
		uint32 y_offs = 0;
		pvr_poly_compile(&screen_poly, &txr_ex_ctx);
#ifdef USE_64_BIT_TXR_XFER
		unsigned short *dst = (unsigned short *)txr_base_ex;
		unsigned char *src = &screen[0];

		for( y_offs = 0; y_offs<240; y_offs++ )
		{
			texture_memcpy64_pal(dst, src);
			src += 512;
			dst += 512;
		}
#else
		uint32 x_offs = 0;
		for(y_offs = 0; y_offs < vdp_vislines; y_offs++)
		{
			for(x_offs = 0; x_offs < u_max; x_offs++)
			{
				screen_ex[(y_offs*512)+x_offs] = (pals[screen[(y_offs*512)+x_offs]]);
			}
		}
#ifdef USE_TWIDDLED_TXR
		pvr_txr_load_twiddle(&screen_ex[0], txr_base_ex, 512, 512);
#else
		memcpy4(txr_base_ex, &screen_ex[0], 512*512*2);
#endif // USE_TWIDDLED_TXR
#endif // USE_64_BIT_TXR_XFER
	}
	pvr_wait_ready();
	pvr_scene_begin();
	
	pvr_list_begin(PVR_LIST_OP_POLY);
	pvr_dummy_poly(PVR_LIST_OP_POLY);
	pvr_list_finish();

	pvr_list_begin(PVR_LIST_TR_POLY);
	pvr_prim(&screen_poly, sizeof(screen_poly));

	vert.flags = PVR_CMD_VERTEX;
	vert.argb = PVR_PACK_COLOR(1,1,1,1);
	vert.oargb = PVR_PACK_COLOR(1,0,0,0);

	vert.z = 256.0f;
	vert.x = x;
	vert.y = y + y_size;
	vert.u = 0;
	vert.v = vdp_vislines/512.0f;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x; 
	vert.y = y;
	vert.u = 0;
	vert.v = 0;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x + x_size;
	vert.y = y + y_size;
	vert.u = u_max/512.0f;
	vert.v = vdp_vislines/512.0f;
	pvr_prim(&vert, sizeof(vert));
	
	vert.flags = PVR_CMD_VERTEX_EOL;
	vert.x = x + x_size;
	vert.y = y;
	vert.u = u_max/512.0f;
	vert.v = 0;
	pvr_prim(&vert, sizeof(vert));

	pvr_list_finish();

	pvr_scene_finish();
}

/*** logging functions ***/
void ui_log_debug3(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 7) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("DEBG (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_debug2(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 6) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("DEBG (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_debug1(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 5) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("DEBG (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_user(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 4) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("USER (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_verbose(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 3) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("---- (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_normal(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 2) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("---- (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_critical(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel >= 1) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("CRIT (%05X) %s\n", cpu68k_clocks, msg);
  }
}

void ui_log_request(const char *text, ...)
{
  va_list ap;
  char msg[512];
  if (gen_loglevel) {
    va_start(ap, text);
    vsprintf((char *)msg, text, ap);
    va_end(ap);
    printf("---- (%05X) %s\n", cpu68k_clocks, msg);
  }
}

/*** ui_err - log error message and quit ***/

void ui_err(const char *text, ...)
{
  va_list ap;
  char msg[512];

  va_start(ap, text);
  vsprintf((char *)msg, text, ap);
  va_end(ap);
  while(!ui_check_cond(CONT_B))
  {
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);
	pvr_dummy_poly(PVR_LIST_OP_POLY);
	pvr_list_finish();
	pvr_list_begin(PVR_LIST_TR_POLY);
	draw_poly_strf(40.0f, 224.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "%s",msg);
	draw_poly_strf(40.0f, 400.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "PRESS B TO RETURN TO MENU.");
	pvr_list_finish();
	pvr_scene_finish();
  }
  gen_quit = 1;
}

void ui_musiclog(uint8 *data, unsigned int length)
{
}
